@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2015-2017 Phusion Holding B.V.
+ *  Copyright (c) 2015-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -27,18 +27,12 @@
 #define _PASSENGER_API_SERVER_UTILS_H_
 
 /**
- * Utility code shared by agent/Core/ApiServer.h, agent/UstRouter/ApiServer.h
- * and agent/Watchdog/ApiServer.h. This code handles authentication and authorization
+ * Utility code shared by agent/Core/ApiServer.h and agent/Watchdog/ApiServer.h.
+ * This code handles authentication and authorization
  * of connected ApiServer clients.
  *
  * This file consists of the following items.
  *
- * ## API accounts
- *
- * API servers can be password protected. They support multiple accounts,
- * each with its own privilege level. These accounts are represented by
- * ApiAccount, stored in ApiAccountDatabase objects.
- *
  * ## Authorization
  *
  * The authorizeXXX() family of functions implement authorization checking on a
@@ -56,7 +50,7 @@
  */
 
 #include <boost/foreach.hpp>
-#include <boost/bind.hpp>
+#include <boost/bind/bind.hpp>
 #include <boost/regex.hpp>
 #include <oxt/macros.hpp>
 #include <oxt/backtrace.hpp>
@@ -64,6 +58,7 @@
 #include <sys/types.h>
 #include <string>
 #include <vector>
+#include <stdexcept>
 #include <cstddef>
 #include <cstring>
 #include <jsoncpp/json.h>
@@ -72,14 +67,16 @@
 #include <Exceptions.h>
 #include <DataStructures/LString.h>
 #include <DataStructures/StringKeyTable.h>
+#include <ServerKit/llhttp.h>
 #include <ServerKit/Server.h>
 #include <ServerKit/HeaderTable.h>
+#include <LoggingKit/LoggingKit.h>
+#include <LoggingKit/Context.h>
 #include <Utils.h>
-#include <Utils/IOUtils.h>
-#include <Utils/BufferedIO.h>
-#include <Utils/StrIntUtils.h>
+#include <IOTools/IOUtils.h>
+#include <IOTools/BufferedIO.h>
+#include <StrIntTools/StrIntUtils.h>
 #include <Utils/VariantMap.h>
-#include <Core/ApplicationPool/Pool.h>
 #include <Shared/ApplicationPoolApiKey.h>
 
 namespace Passenger {
@@ -91,102 +88,6 @@ using namespace std;
 inline string truncateApiKey(const StaticString &apiKey);
 
 
-/*******************************
- *
- * API accounts
- *
- *******************************/
-
-struct ApiAccount {
-	string username;
-	string password;
-	bool readonly;
-};
-
-class ApiAccountDatabase {
-private:
-	vector<ApiAccount> database;
-
-	bool levelDescriptionIsReadOnly(const StaticString &level) const {
-		if (level == "readonly") {
-			return true;
-		} else if (level == "full") {
-			return false;
-		} else {
-			throw ArgumentException("Invalid privilege level " + level);
-		}
-	}
-
-public:
-	/**
-	 * Add an account to the database with the given parameters.
-	 *
-	 * @throws ArgumentException One if the input arguments contain a disallowed value.
-	 */
-	void add(const string &username, const string &password, bool readonly) {
-		if (OXT_UNLIKELY(username == "api")) {
-			throw ArgumentException("It is not allowed to register an API account with username 'api'");
-		}
-
-		ApiAccount account;
-		account.username = username;
-		account.password = password;
-		account.readonly = readonly;
-		database.push_back(account);
-	}
-
-	/**
-	 * Add an account to the database. The account parameters are determined
-	 * by a description string in the form of [LEVEL]:USERNAME:PASSWORDFILE.
-	 * LEVEL is one of:
-	 *
-	 *   readonly    Read-only access
-     *   full        Full access (default)
-	 *
-	 * @throws ArgumentException One if the input arguments contain a disallowed value.
-	 */
-	void add(const StaticString &description) {
-		ApiAccount account;
-		vector<string> args;
-
-		split(description, ':', args);
-
-		if (args.size() == 2) {
-			account.username = args[0];
-			account.password = strip(readAll(args[1]));
-			account.readonly = false;
-		} else if (args.size() == 3) {
-			account.username = args[1];
-			account.password = strip(readAll(args[2]));
-			account.readonly = levelDescriptionIsReadOnly(args[0]);
-		} else {
-			throw ArgumentException("Invalid authorization description '" + description + "'");
-		}
-
-		if (OXT_UNLIKELY(account.username == "api")) {
-			throw ArgumentException("It is not allowed to register an API account with username 'api'");
-		}
-		database.push_back(account);
-	}
-
-	bool empty() const {
-		return database.empty();
-	}
-
-	const ApiAccount *lookup(const StaticString &username) const {
-		vector<ApiAccount>::const_iterator it, end = database.end();
-
-		for (it = database.begin(); it != end; it++) {
-			if (it->username == username) {
-				return &(*it);
-			}
-		}
-
-		return NULL;
-	}
-};
-
-
 /*******************************
  *
  * Authorization functions
@@ -265,7 +166,7 @@ authorize(ApiServer *server, Client *client, Request *req) {
 		}
 	}
 
-	if (server->apiAccountDatabase->empty()) {
+	if (server->getApiAccountDatabase().empty()) {
 		SKC_INFO_FROM_STATIC(server, client,
 			"Authenticated as administrator because API account database is empty");
 		auth.apiKey = ApplicationPool2::ApiKey::makeSuper();
@@ -286,7 +187,8 @@ authorize(ApiServer *server, Client *client, Request *req) {
 				auth.canModifyPool = true;
 			}
 		} else {
-			const ApiAccount *account = server->apiAccountDatabase->lookup(username);
+			const typename ApiServer::ApiAccount *account =
+				server->getApiAccountDatabase().lookup(username);
 			if (account != NULL && constantTimeCompare(password, account->password)) {
 				SKC_INFO_FROM_STATIC(server, client,
 					"Authenticated with administrator account: " << username);
@@ -384,7 +286,7 @@ struct ApiServerInternalHttpRequest {
 	Request *req;
 
 	string address;
-	http_method method;
+	llhttp_method_t method;
 	string uri;
 	StringKeyTable<string> headers;
 	boost::function<void (ApiServerInternalHttpResponse<Server, Client, Request>)> callback;
@@ -474,7 +376,7 @@ apiServerMakeInternalHttpRequestThreadMain(ApiServerInternalHttpRequest<Server,
 		BufferedIO io(conn);
 
 		string header;
-		header.append(http_method_str(req.method));
+		header.append(llhttp_method_name(req.method));
 		header.append(" ", 1);
 		header.append(req.uri);
 		header.append(" HTTP/1.1\r\n");
@@ -559,7 +461,7 @@ apiServerMakeInternalHttpRequestThreadMain(ApiServerInternalHttpRequest<Server,
 			}
 		}
 
-		if (req.bodyProcessor != NULL) {
+		if (req.bodyProcessor) {
 			req.bodyProcessor(req, resp, io);
 		} else {
 			resp.body = io.readAll(&req.timeout);
@@ -763,12 +665,11 @@ apiServerProcessReopenLogs(Server *server, Client *client, Request *req) {
 	if (req->method != HTTP_POST) {
 		apiServerRespondWith405(server, client, req);
 	} else if (authorizeAdminOperation(server, client, req)) {
-		int e;
 		ServerKit::HeaderTable headers;
 		headers.insert(req->pool, "Content-Type", "application/json");
 
-		string logFile = getLogFile();
-		if (logFile.empty()) {
+		ConfigKit::Store config = LoggingKit::context->getConfig();
+		if (!config["target"].isMember("path")) {
 			server->writeSimpleResponse(client, 500, &headers, "{ \"status\": \"error\", "
 				"\"code\": \"NO_LOG_FILE\", "
 				"\"message\": \"" PROGRAM_NAME " was not configured with a log file.\" }\n");
@@ -778,40 +679,61 @@ apiServerProcessReopenLogs(Server *server, Client *client, Request *req) {
 			return;
 		}
 
-		if (!setLogFile(logFile, &e)) {
-			unsigned int bufsize = 1024;
+		LoggingKit::ConfigChangeRequest configReq;
+		vector<ConfigKit::Error> errors;
+		bool ok;
+		try {
+			// We deliberately ignore the target.stderr key.
+			// If the log file was equal to stderr then we'll want
+			// to reopen the log file anyway.
+			Json::Value updates;
+			updates["target"] = config["target"]["path"];
+			ok = LoggingKit::context->prepareConfigChange(updates,
+				errors, configReq);
+		} catch (const SystemException &e) {
+			unsigned int bufsize = 2048;
 			char *message = (char *) psg_pnalloc(req->pool, bufsize);
 			snprintf(message, bufsize, "{ \"status\": \"error\", "
-				"\"code\": \"LOG_FILE_OPEN_ERROR\", "
-				"\"message\": \"Cannot reopen log file %s: %s (errno=%d)\" }",
-				logFile.c_str(), strerror(e), e);
+				"\"code\": \"OS_ERROR\", "
+				"\"message\": \"Cannot reopen log files: %s\" }",
+				e.what());
+			server->writeSimpleResponse(client, 500, &headers, message);
+			if (!req->ended()) {
+				server->endRequest(&client, &req);
+			}
+			return;
+		} catch (const std::exception &e) {
+			unsigned int bufsize = 2048;
+			char *message = (char *) psg_pnalloc(req->pool, bufsize);
+			snprintf(message, bufsize, "{ \"status\": \"error\", "
+				"\"code\": \"GENERIC_ERROR\", "
+				"\"message\": \"Cannot reopen log files: %s\" }",
+				e.what());
 			server->writeSimpleResponse(client, 500, &headers, message);
 			if (!req->ended()) {
 				server->endRequest(&client, &req);
 			}
 			return;
 		}
-		P_NOTICE("Log file reopened.");
-
-		if (hasFileDescriptorLogFile()) {
-			if (!setFileDescriptorLogFile(getFileDescriptorLogFile(), &e)) {
-				unsigned int bufsize = 1024;
-				char *message = (char *) psg_pnalloc(req->pool, bufsize);
-				snprintf(message, bufsize, "{ \"status\": \"error\", "
-					"\"code\": \"FD_LOG_FILE_OPEN_ERROR\", "
-					"\"message\": \"Cannot reopen file descriptor log file %s: %s (errno=%d)\" }",
-					getFileDescriptorLogFile().c_str(), strerror(e), e);
-				server->writeSimpleResponse(client, 500, &headers, message);
-				if (!req->ended()) {
-					server->endRequest(&client, &req);
-				}
-				return;
+		if (!ok) {
+			unsigned int bufsize = 2048;
+			char *message = (char *) psg_pnalloc(req->pool, bufsize);
+			snprintf(message, bufsize, "{ \"status\": \"error\", "
+				"\"code\": \"CONFIG_VALIDATION_ERROR\", "
+				"\"message\": \"Cannot reopen log files:"
+					" invalid logging system configuration: %s\" }",
+				ConfigKit::toString(errors).c_str());
+			server->writeSimpleResponse(client, 500, &headers, message);
+			if (!req->ended()) {
+				server->endRequest(&client, &req);
 			}
-			P_NOTICE("File descriptor log file reopened.");
+			return;
 		}
 
-		server->writeSimpleResponse(client, 200, &headers, "{ \"status\": \"ok\" }\n");
+		LoggingKit::context->commitConfigChange(configReq);
+		P_NOTICE("All log file(s) reopened.");
 
+		server->writeSimpleResponse(client, 200, &headers, "{ \"status\": \"ok\" }\n");
 		if (!req->ended()) {
 			server->endRequest(&client, &req);
 		}
@@ -838,8 +760,37 @@ _apiServerProcessReinheritLogsResponseBody(
 	}
 
 	int fd = readFileDescriptorWithNegotiation(io.getFd(), &req.timeout);
-	setLogFileWithFd(logFilePath, fd);
-	safelyClose(fd);
+	FdGuard guard(fd, __FILE__, __LINE__);
+	P_LOG_FILE_DESCRIPTOR_PURPOSE(fd, "Reinherited log file handle");
+
+	ConfigKit::Store oldConfig = LoggingKit::context->getConfig();
+	Json::Value config;
+	vector<ConfigKit::Error> errors;
+	LoggingKit::ConfigChangeRequest configReq;
+	bool ok;
+
+	config["target"] = oldConfig["target"];
+	config["target"]["fd"] = fd;
+	config["target"].removeMember("stderr");
+	try {
+		ok = LoggingKit::context->prepareConfigChange(config,
+			errors, configReq);
+	} catch (const std::exception &e) {
+		resp.status = InternalResponse::ERROR_INTERNAL;
+		resp.errorLogs.append("Error reconfiguring logging system: ");
+		resp.errorLogs.append(e.what());
+		return;
+	}
+	if (!ok) {
+		resp.status = InternalResponse::ERROR_INTERNAL;
+		resp.errorLogs.append("Error reconfiguring logging system: ");
+		resp.errorLogs.append(ConfigKit::toString(errors));
+		return;
+	}
+
+	LoggingKit::context->commitConfigChange(configReq);
+	guard.clear();
+	P_NOTICE("All log file(s) reinherited.");
 }
 
 template<typename Server, typename Client, typename Request>
